/*
 * This file is part of SionTower.
 *
 * 
 * David Saltares M&aacute;rquez (C) 2011
 * <david.saltares@gmail.com>
 *
 * 
 * SionTower examples are free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License ad
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) ant later version.
 *
 * SionTower examples are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SionTower examples.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SIONTOWER_TRUNK_SRC_INCLUDE_SOUND3D_H_
#define SIONTOWER_TRUNK_SRC_INCLUDE_SOUND3D_H_

#include <vector>

#include <OGRE/Ogre.h>

#include "soundFX.h"

//! Clase para simular sonidos en el espacio con una fuente y una entidad que escucha

/**
 *  @author David Saltares M&aacute;rquez
 *  @date 29-12-2010
 *
 *  Esta clase utiliza el recurso SoundFX o el gestor SoundFXManager para
 *  reproducir efectos de sonido con un efecto pseudo-3D. Sound3D asocia un
 *  nodo emisor de sonido, un nodo que escucha y un efecto a reproducir. El
 *  volumen de la reproducci&oacute;n var&iacute;a en funci&oacute;n de la distancia y se produce un
 *  efecto est&eacute;reo en los altavoces en funci&oacute;n del &aacute;ngulo entre los nodos.
 *
 *  Es necesario actualizar el Sound3D en cada iteraci&oacute;n del bucle de juego
 *  para que el volumen y el efecto est&eacute;reo reflejen lo que ocurre en la escena.
 *
 *  Ejemplo:
 *
 *  \code
 *  Sound3D grito("disparo.wav");
 *  grito.setListenerNode(nodoPersonaje);
 *  grito.setSourceNode(nodoEnemigo);
 *
 *  ...
 *
 *  while(!salir) {
 *      ...
 *
 *      if(enemigoHerido)
 *          grito.play();
 *      ...
 *
 *      Sound3D::update3DSound();
 *
 *      ...
 *  }
 *
 *  \endcode
 */
class Sound3D {
    public:
        /**
         *  Constructor
         *
         *  @param name nombre del recurso tipo SoundFX a cargar.
         *  @param group nombre del grupo al que pertenece el recurso. Por
         *  defecto es el grupo "General".
         *  @param listenerNode puntero al nodo que escucha el sonido.
         *  @param sourceNode puntero al nodo que emite el sonido.
         */
        Sound3D(const Ogre::String& name,
                const Ogre::String& group = Ogre::ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME,
                Ogre::SceneNode* listenerNode = 0,
                Ogre::SceneNode* sourceNode = 0);

        /**
         *  Constructor
         *
         *  @param s puntero inteligente al sonido asociado al efecto 3D.
         *  @param listenerNode puntero al nodo que escucha el sonido.
         *  @param sourceNode puntero al nodo que emite el sonido.
         */
        Sound3D(SoundFXPtr s,
                Ogre::SceneNode* listenerNode = 0,
                Ogre::SceneNode* sourceNode = 0);

        /**
         *  Destructor
         */
        ~Sound3D();

        /**
         *  @param loop n&uacute;mero de veces que se reproduce el efecto de sonido.
         *  -1 equivale a infinitas repeticiones, 0 equivale a 1 reproducci&oacute;n.
         *  En general, si bucle = n, el efecto se reproducir&aacute; n + 1 veces.
         *
         *  Reproduce el efecto 3D.
         */
        void play(int loop = 0);

        /**
         *  Actualiza el volumen y el efecto est&eacute;reo de todos los efectos 3D en
         *  reproducci&oacute;n. Es necesario llamar a este m&eacute;todo en cada iteraci&oacute;n
         *  del bucle principal.
         */
        static void update3DSound();

        /**
         *  @param channel canal que acaba de finalizar su reproducci&oacute;n.
         *
         *  M&eacute;todo que utiliza el callback para gestionar internamente el fin
         *  de la reproducci&oacute;n de un efecto en un canal.
         */
        static void channelFinished(int channel);

        /**
         *  @return puntero al nodo que escucha el efecto.
         */
        const Ogre::SceneNode* getListenerNode() const;

        /**
         *  @return puntero al nodo que emite el efecto.
         */
        const Ogre::SceneNode* getSourceNode() const;

        /**
         *  @param e puntero al nuevo nodo que escucha el efecto.
         */
        void setListenerNode(const Ogre::SceneNode* listenerNode);

        /**
         *  @param f puntero al nuevo nodo que emite el efecto.
         */
        void setSourceNode(const Ogre::SceneNode* sourceNode);

    private:
        SoundFXPtr _soundFXPtr;
        const Ogre::SceneNode* _listenerNode;
        const Ogre::SceneNode* _sourceNode;
        
        // Vector con los sonidos 3D en reproducci&oacute;n junto a su canal asociado.
        static std::vector<std::pair<int, Sound3D*> > _sounds;
        // Vector con todos los canales de forma que canales[n] indica la
        // posici&oacute;n en sonidos de ese canal.
        static std::vector<int> _channels;
        static int _numUsedChannels;

        void update(int channel);
};

/**
 *  @param channel canal que ha terminado de reproducir un efecto.
 *
 *  Funci&oacute;n de callback para SDL_Mixer que nos avisa cuando un canal ha
 *  terminado de reproducir un efecto.
 */
void channelFinished(int channel);

/**
 *  @mainpage Sistema de audio 3D de Sion Tower
 *
 *  <strong><center><h2>David Saltares M&aacute;rquez</h2></center></strong>
 *  <ul>
 *      <li><strong>Correo electr&oacute;nico</strong>: <a href="mailto:david.saltares@gmail.com">david.saltares@gmail.com</a></li>
 *      <li><strong>Web del proyecto</strong>: <a href="http:://siondream.com/blog/category/proyectos/pfc">http:://siondream.com/blog/category/proyectos/pfc</a></li>
 *      <li><strong>Forja de Red Iris</strong>: <a href="http://forja.rediris.es/projects/cusl5-iberogre">http://forja.rediris.es/projects/cusl5-iberogre</a></li>
 *  </ul>
 *
 *  <h3>Descripci&oacute;n</h3>
 *  
 *  <p>Sion Tower es un videojuego libre de estrategia y acci&oacute;n desarrollado
 *  como Proyecto Fin de Carrera por David Saltares M&aacute;rquez. Est&aacute; muy
 *  relacionado con <a href="http://osl2.uca.es/iberogre">IberOgre</a>,
 *  la wiki de Ogre3D en espa&ntilde;ol (del mismo autor).</p>
 *
 *  <p>El sistema de audio 3D de Sion Tower puede utilizarse de forma muy
 *  sencilla en otros proyectos que empleen Ogre3D como engine de renderizado.
 *  Est&aacute; liberado bajo licencia GPL v3, consulta el fichero LICENSE.txt distribuido
 *  con el proyecto. Utiliza C++ est&aacute;ndar y es compatible con todas las
 *  plataformas que soporten SDL mixer y Ogre3D.</p>
 *
 *  <ul>
 *      <li>Efectos de sonido y m&uacute;sicas: clases SoundFX y Song que permiten cargar ficheros WAV, MOD, MID, OGG y MP3.</li>
 *      <li>Integraci&oacute;n con el sistema de gesti&oacute;n de recursos de Ogre3D con las clases SoundFXManager y SongManager</li>
 *      <li>Efecto de audio 3D en funci&oacute;n de la distancia y &aacute;ngulo entre dos nodos de Ogre3D con la clase Sound3D</li>
 *  </ul>
 *
 *  <h3>Dependencias y uso</h3>
 *
 *  <p>El sistema de audio 3D de Sion Tower depende de las bibliotecas
 *  <strong>Ogre3D</strong> y <strong>SDL Mixer</strong>. En funci&oacute;n de tu sistema
 *  deber&aacute;s instalarlas de manera diferente. Para incorporar este sistema de
 *  audio 3D a tu proyecto s&oacute;lo debes a&ntilde;adir su c&oacute;digo fuente. Por supuesto,
 *  debes asegurarte que tu sistema de compilaci&oacute;n tenga en cuenta y enlace
 *  dicho c&oacute;digo (configura tu makefile o proyecto de IDE).</p>
 */
    


#endif  // SIONTOWER_TRUNK_SRC_INCLUDE_SOUND3D_H_
